/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"

#include "selectB.h"
#include "registers.h"
#include "ds485.h"
#include "datamodel.h"
#include "mmi.h"
#include "selectA.h"
#include "bsp.h"
#include "dsmapi.h"
#include "util.h"
#include "device-api-proto.h"
#include "vdc/vdc-api-proto.h"

#include <syslog.h>

Q_DEFINE_THIS_MODULE(selectB)

enum {
    SELECT_B_WAIT_FOR_DSID_TICKS            = BSP_TICKS(5000),
    SELECT_B_WAIT_FOR_PARAM_TICKS           = BSP_TICKS(1250),
    SELECT_B_WAIT_FOR_RAMSTATS_TICKS        = BSP_TICKS(1250),
    SELECT_B_WAIT_AFTER_READY               = BSP_TICKS(2000),
    SELECT_B_TX_TIMEOUT_TICKS               = BSP_TICKS(10000),
    SELECT_B_GET_FIRST_PING_DEVICE_TICKS    = BSP_TICKS(60000),     /**< re-check after 1 minute */
    SELECT_B_ONE_MINUTE_TICKS               = BSP_TICKS(60000),
    SELECT_B_ONE_SECOND_TICKS               = BSP_TICKS(1000),
    SELECT_B_MINUTES_PER_24_HOURS           = (24 * 60),
};

typedef struct {
    DSLinkOperation_t operation;
    uint8_t value;
} SelectBdSLinkInfo;

typedef struct {
    QEvent super;
    char dsuid[DSUID_STR_LEN];
    char vdcdsuid[DSUID_STR_LEN];
    uint32_t message_id;
} SelectBAnnounceDeviceEvt;

/* Active object class -----------------------------------------------------*/
typedef struct {
    QActive super;                          /**< active object base class */
    QTimeEvt timeEvt;                       /**< timer */
    QTimeEvt txTimer;                       /**< transmission timer */
    QTimeEvt pingTimer;                     /**< ping timer */
    QTimeEvt minuteTimer;                   /**< minute timer */
    uint32_t minuteCount;                   /**< minute count since dsm start */
    QTimeEvtCtr defaultAutoPingInterval;    /**< default auto ping interval time in milliseconds */
    QTimeEvtCtr defaultOneTimePingInterval; /**< default one-time ping interval time in milliseconds */
    QTimeEvtCtr autoPingInterval;           /**< automatic ping interval time in milliseconds */
    QTimeEvtCtr oneTimePingInterval;        /**< one-time ping interval time in milliseconds */
    QTimeEvtCtr currentPingTicks;           /**< current ping interval time in ticks */
    QEQueue deferredCommandQueue;           /**< deferred command event queue */
    QEQueue deferredDsidQueue;              /**< deferred dSID event queue */
    QEQueue deferredInactiveDevToPingQueue; /**< deferred inactive device to ping event queue */
    QStateHandler returnState;              /**< return state */
    QStateHandler superState;               /**< super state */
    QStateHandler returnStateSendDsid;      /**< */
    QStateHandler superStateSendDsid;       /**< */
    QStateHandler paramOperation;           /**< */
    QStateHandler selectNextState;          /**< state after a successful SelectB */
    QStateHandler selectTopState;           /**< top-state for SelectB call */
    uint32_t dsmid;                         /**< dSM ID (only serial number) */
    uint32_t sendingDsmid;                  /**< dSM ID to send to devices (only serial number) */
    char dsuid[DSUID_STR_LEN];              /**< selected dSUID */
    char vdc_dsuid[DSUID_STR_LEN];          /**< selected vdc dSUID */
    uint32_t selectedDsid;                  /**< selected dSID */
    uint8_t selectedParamBank;              /**< selected parameter bank of the device */
    uint8_t paramBank;                      /**< cached parameter block for get/set parameter */
    uint8_t paramOffset;                    /**< cached parameter offset for get/set parameter */
    uint8_t paramValue;                     /**< cached parameter value for set parameter */
    uint32_t pingSingleDeviceDsid;          /**< dSID of single device to ping */
    uint16_t pingDeviceId;                  /**< device ID to ping */
    uint16_t pingSingleDeviceId;            /**< single device ID to ping */
    uint16_t oneTimePingInitialDeviceId;    /**< initial device ID for one-time ping */
    uint8_t pingMode;                       /**< ping mode */
    uint8_t pingCount;                      /**< ping count */
    uint8_t pingSingleDeviceCount;          /**< ping single device count */
    bool pingSingleTriggeredByInvalid;      /**< single device ping was initiated by invalid response */
    bool pingTimerElapsed;                  /**< ping timer elapsed flag */
    uint8_t sensorPushType;                 /**< sensor type to be pushed to the devices */
    uint16_t sensorPushValue;               /**< value to be pushed to the devies */
    uint16_t deviceId;                      /**< device ID */
    uint16_t fid;                           /**< cached function ID */
    uint16_t ver;                           /**< cached firmware version */
    uint16_t vid;                           /**< cached vendor ID */
    uint8_t localZoneId;                    /**< local zone ID */
    uint16_t zoneId;                        /**< zone ID (global) */
    uint8_t groupMembership[8];             /**< group membership */
    uint8_t groupId;                        /**< group ID */
    uint8_t apartmentId;                    /**< cached apartment ID */
    uint8_t deviceAddress;                  /**< cached dS485 device address */
    bool helloReceived;                     /**< hello telegram received flag */
    uint8_t transactionId;                  /**< last transaction ID */
    uint8_t jumbleRound;                    /**< jumble round count */
    uint8_t jumbleDeviceId;                 /**< current jumble device ID */
    bool deviceRegistered;                  /**< device registered flag */
    bool setDsmId;                          /**< set dSM ID (i.e. dS485 circuit/apartment ID) flag */
    uint8_t timeEvtArmed;                   /**< timer is armed */
    bool configReady;                       /**< configuration ready flag */
    bool jumbleDisabled;                    /**< jumble disabled flag */
    uint8_t getStatisticsParamOffset;       /**< cached parameter offset for get statistics */
    uint16_t getStatisticsDeviceId;         /**< cached device ID for get statistics */
    uint8_t getSensorId;                    /**< cached sensor id */
    SelectBdSLinkInfo dSLinkInfo;           /**< cache for dSLink operations. */
    bool dSLinkDeviceSelect;                /**< dSLink device select flag */
    bool deviceSelectionReset;              /**< current device selection was reset */
    bool pingEvery24HoursEnabled;           /**< ping every 24 hours */
    bool clearPingDeviceId;                 /**< clear ping device ID flag */
    bool deviceIsKnown;                     /**< device is already known in datamodel */
    uint16_t selectSelection;               /**< device/group selection for the SeletB call */
    BinaryInputConfig binaryInputTable[BINARY_INPUT_TABLE_LENGTH];  /**< the binary input table */
    SensorConfig sensorTable[SENSOR_TABLE_LENGTH];                  /**< the sensor table */
    OpcConfig opcTable[OPC_TABLE_LENGTH];   /**< the opc table */
    bool returnToIdleState;                 /**< when a ping configuration change is received during ping round then return to idle state */
    uint8_t OPCchannel;
    uint16_t OPCvalue;
    uint8_t OPCscene;
    uint8_t OPCoperation;
    bool OPCis16bit;
    uint32_t messageId;
    uint8_t selectedScene;                  /**< selected scene table entry of the device */
    uint8_t extraCommand;                   /**< cached extra command */
    bool extraCommandDeviceSelect;          /**< extra command device select flag */
} SelectB;

/* Local objects -----------------------------------------------------------*/
static SelectB l_selectB;   /* the single instance of the downstream selectB active object */
static QEvent const * l_deferredCommandQueueSto[128];            /* deferred queue storage */
static QEvent const * l_deferredDsidQueueSto[128];                /* deferred queue storage */
static QEvent const * l_deferredInactiveDevToPinqQueueSto[128];   /* deferred queue storage */

/* Global-scope objects ----------------------------------------------------*/
QActive * const AO_SelectB = (QActive *)&l_selectB;      /* "opaque" AO pointer */

/* Local methods -----------------------------------------------------------*/
static QState SelectB_initial(SelectB *me, QEvent const *e);
static QState SelectB_wait_for_ready(SelectB *me, QEvent const *e);

static QState SelectB_top(SelectB *me, QEvent const *e);

static QState SelectB_idle(SelectB *me, QEvent const *e);

static QState SelectB_busy(SelectB *me, QEvent const *e);
static QState SelectB_register_lookup_device(SelectB *me, QEvent const *e);
static QState SelectB_register_get_params(SelectB *me, QEvent const *e);
static QState SelectB_register_get_sensors(SelectB *me, QEvent const *e);
static QState SelectB_register_get_binary_inputs(SelectB *me, QEvent const *e);
static QState SelectB_register_get_LTMode_param(SelectB *me, QEvent const *e);
static QState SelectB_register_get_Mode_param(SelectB *me, QEvent const *e);
static QState SelectB_register_get_group_membership(SelectB *me, QEvent const *e);
static QState SelectB_register_get_Opc_param(SelectB *me, QEvent const *e);
static QState SelectB_register_get_zone(SelectB *me, QEvent const *e);
static QState SelectB_register_get_name(SelectB *me, QEvent const *e);
static QState SelectB_user_protobuf_message_request(SelectB *me, QEvent const *e);

static QState SelectB_ping_busy(SelectB *me, QEvent const *e);
static QState SelectB_ping_get_device_info(SelectB *me, QEvent const *e);
static QState SelectB_ping_send_ping(SelectB *me, QEvent const *e);
static QState SelectB_ping_get_next_device(SelectB *me, QEvent const *e);

static QState SelectB_ping_single_device(SelectB *me, QEvent const *e);
static QState SelectB_stopped(SelectB *me, QEvent const *e);


static void SelectB_SendDsmDoneEvent(DeviceParameterEvt const *reqEvt, uint8_t error);

/*..........................................................................*/
void SelectB_ctor(void)
{
    SelectB *me = &l_selectB;
    /* call active object (superclass) constructor */
    QActive_ctor(&me->super, (QStateHandler)&SelectB_initial);
    /* initialize time events */
    QTimeEvt_ctor(&me->timeEvt, SELECT_B_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->txTimer, SELECT_B_TX_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->pingTimer, SELECT_B_PING_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->minuteTimer, SELECT_B_MINUTE_TIMEOUT_SIG);
    /* initialize deferred queues */
    QEQueue_init(&me->deferredCommandQueue, l_deferredCommandQueueSto, Q_DIM(l_deferredCommandQueueSto), "command", &me->super);
    QEQueue_init(&me->deferredDsidQueue, l_deferredDsidQueueSto, Q_DIM(l_deferredDsidQueueSto), "dsid", &me->super);
    QEQueue_init(&me->deferredInactiveDevToPingQueue, l_deferredInactiveDevToPinqQueueSto, Q_DIM(l_deferredInactiveDevToPinqQueueSto), "inactiveDevToPing", &me->super);
    /* initialize extended state variables */
    me->apartmentId = 0;
    me->deviceAddress = DS485_INVALID_ADDRESS;
    me->setDsmId = FALSE;
    me->helloReceived = FALSE;
    me->timeEvtArmed = FALSE;
    me->pingCount = 0;
    me->pingDeviceId = 0;
    me->oneTimePingInitialDeviceId = 0;
    me->configReady = FALSE;
    me->jumbleDisabled = FALSE;
    me->deviceSelectionReset = FALSE;
    me->pingTimerElapsed = FALSE;
    me->pingEvery24HoursEnabled = FALSE;
    me->minuteCount = 0;
    me->clearPingDeviceId = FALSE;
    me->returnToIdleState = FALSE;
    memset(me->binaryInputTable, 0, sizeof(me->binaryInputTable));
    memset(me->sensorTable, 0, sizeof(me->sensorTable));
}

/*..........................................................................*/
uint8_t SelectB_GetDeferredCommandQueueMargin(void)
{
    return QEQueue_getNFree(&l_selectB.deferredCommandQueue);
}

/*..........................................................................*/
uint8_t SelectB_GetDeferredDsidQueueMargin(void)
{
    return QEQueue_getNFree(&l_selectB.deferredDsidQueue);
}

/*..........................................................................*/
static QState SelectB_initial(SelectB *me, QEvent const *e __attribute__((unused)))
{
    /* QS dictionary entries */
    QS_OBJ_DICTIONARY(&l_selectB);

    QS_SIG_DICTIONARY(SELECT_B_SET_DSM_ID_SIG, 0);
    QS_SIG_DICTIONARY(SELECT_B_HELLO_SIG, 0);
    QS_SIG_DICTIONARY(SELECT_B_GET_DEVICE_STATISTICS_SIG, 0);
    QS_SIG_DICTIONARY(SELECT_B_GET_DEVICE_SENSOR_SIG, 0);
    QS_SIG_DICTIONARY(SELECT_B_GET_DEVICE_PARAMETER_SIG, 0);
    QS_SIG_DICTIONARY(SELECT_B_SET_DEVICE_PARAMETER_SIG, 0);
    QS_SIG_DICTIONARY(SELECT_B_TEST_TRANSMISSION_QUALITY_SIG, 0);
    QS_SIG_DICTIONARY(SELECT_B_GET_DEVICE_DSLINK_SIG, 0);
    QS_SIG_DICTIONARY(SELECT_B_GET_ZONE_GROUP_DSLINK_SIG, 0);
    QS_SIG_DICTIONARY(SELECT_B_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(SELECT_B_PING_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(SELECT_B_TX_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(SELECT_B_RESTART_REGISTRATION, 0);

    QS_FUN_DICTIONARY(&SelectB_wait_for_ready);

    QS_FUN_DICTIONARY(&SelectB_top);

    QS_FUN_DICTIONARY(&SelectB_idle);

    QS_FUN_DICTIONARY(&SelectB_busy);

    QS_FUN_DICTIONARY(&SelectB_ping_select);
    QS_FUN_DICTIONARY(&SelectB_ping_send_ping);
    QS_FUN_DICTIONARY(&SelectB_ping_get_next_device);

    QS_FUN_DICTIONARY(&SelectB_register);
    QS_FUN_DICTIONARY(&SelectB_register_select_dsid);
    QS_FUN_DICTIONARY(&SelectB_register_wait);
    QS_FUN_DICTIONARY(&SelectB_register_retrieve_device);
    QS_FUN_DICTIONARY(&SelectB_register_get_params);
    QS_FUN_DICTIONARY(&SelectB_register_add_device);
    QS_FUN_DICTIONARY(&SelectB_register_set_device_id);
    QS_FUN_DICTIONARY(&SelectB_register_set_local_zone_id);

    QS_FUN_DICTIONARY(&SelectB_set_dsm_id);

    QS_FUN_DICTIONARY(&SelectB_change_zone_on_device);
    QS_FUN_DICTIONARY(&SelectB_change_zone_on_device_select);
    QS_FUN_DICTIONARY(&SelectB_change_zone_on_device_set_local_zone_id);

    QS_FUN_DICTIONARY(&SelectB_dev_statistics);
    QS_FUN_DICTIONARY(&SelectB_dev_statistics_get_ramstats);

    QS_FUN_DICTIONARY(&SelectB_stopped);

    QS_FUN_DICTIONARY(&SelectB_opc_config);
    QS_FUN_DICTIONARY(&SelectB_register_get_Opc_param);
    QS_FUN_DICTIONARY(&SelectB_opc_parameter_set);

    /* subscribe events */
    QActive_subscribe((QActive *)me, CONFIG_INIT_SIG);
    QActive_subscribe((QActive *)me, DS485_ADDRESS_SIG);
    QActive_subscribe((QActive *)me, SENSOR_PUSH_VALUE_SIG);

    return Q_TRAN(&SelectB_wait_for_ready);
}

/*..........................................................................*/
static QState SelectB_wait_for_ready(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_EXIT_SIG: {
        /* subscribe events */
        QActive_subscribe((QActive *)me, LONG_UPSTREAM_SIG);
        /* start minute timer */
        QTimeEvt_postEvery(&me->minuteTimer, (QActive *)me, SELECT_B_ONE_MINUTE_TICKS);
        /* start ping timer */
        if (me->pingMode != CIRCUIT_PING_DEVICES_OFF) {
            me->currentPingTicks = me->oneTimePingInterval / BSP_MS_PER_TICKS;
            QTimeEvt_postIn(&me->pingTimer, (QActive *)me, me->currentPingTicks);
        }
        return Q_HANDLED();
    }
    case CONFIG_INIT_SIG: {
        /* cache dSM ID (only serial number) */
        DatamodelConfigEvt const * config = (DatamodelConfigEvt const *)e;
        me->dsmid = config->dsmid.id[8];
        me->dsmid <<= 8;
        me->dsmid |= config->dsmid.id[9];
        me->dsmid <<= 8;
        me->dsmid |= config->dsmid.id[10];
        me->dsmid <<= 8;
        me->dsmid |= config->dsmid.id[11];
        /* cache apartment ID */
        me->apartmentId = config->apartmentId;
        /* cache ping intervals */
        me->defaultAutoPingInterval = config->autoPingInterval;
        me->autoPingInterval = config->autoPingInterval;
        me->defaultOneTimePingInterval = config->oneTimePingInterval;
        me->oneTimePingInterval = config->oneTimePingInterval;
        /* cache ping mode */
        switch (config->dsmInfoFlags & PING_MODE_MASK) {
        case ONE_TIME_OFF_24H_PING_MODE: {
            /* set ping every 24 hours flag */
            me->pingEvery24HoursEnabled = TRUE;
            /* fall through */
        }
        case ONE_TIME_OFF_PING_MODE: {
            /* one-time off ping mode */
            me->pingMode = CIRCUIT_PING_DEVICES_ONE_TIME_OFF;
            break;
        }
        case ONE_TIME_AUTO_PING_MODE: {
            /* one-time automatic ping mode */
            me->pingMode = CIRCUIT_PING_DEVICES_ONE_TIME_AUTOMATIC;
            break;
        }
        default: {
            /* off */
            me->pingMode = CIRCUIT_PING_DEVICES_OFF;
            break;
        }
        }
        /* set configuration ready flag */
        me->configReady = TRUE;
        return Q_TRAN(&SelectB_idle);
    }
    }
    return Q_SUPER(&SelectB_busy);
}

/*..........................................................................*/
static QState SelectB_top(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case DS485_ADDRESS_SIG: {
        me->deviceAddress = ((DS485AddressEvt * const)e)->deviceAddress;
        /* set dSM ID flag */
        me->setDsmId = TRUE;
        return Q_HANDLED();
    }
    case LONG_UPSTREAM_SIG: {
        LongUpstreamEvt * longUpEvt = (LongUpstreamEvt *)e;
        longUpEvt->circuit = calculateCircuitId(me->deviceAddress, me->apartmentId);
        DsmApi_SendLongUpstreamEvent(longUpEvt, FALSE);
        /* create new event */
        LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, SELECT_B_FILTERED_LONG_UPSTREAM_SIG);
        /* copy data to new event */
        filteredUpEvt->flags = longUpEvt->flags;
        filteredUpEvt->circuit = longUpEvt->circuit;
        filteredUpEvt->data = longUpEvt->data;
        filteredUpEvt->mainqualifier = longUpEvt->mainqualifier;
        filteredUpEvt->quality = longUpEvt->quality;
        filteredUpEvt->source = longUpEvt->source;
        filteredUpEvt->subqualifier = longUpEvt->subqualifier;
        /* post event */
        QActive_postLIFO((QActive *)me, (QEvent*)filteredUpEvt);
        return Q_HANDLED();
    }
    case SELECT_B_FILTERED_LONG_UPSTREAM_SIG: {
        LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e;
        if (longUpEvt->mainqualifier == UpstreamMainqualifierStatisticValuesL) {
            /* send statistics event, offset < 0x40 */
            DsmApi_SendDeviceStatisticsEvent(longUpEvt->source, longUpEvt->subqualifier, longUpEvt->data);
        } else if (longUpEvt->mainqualifier == UpstreamMainqualifierStatisticValuesH) {
            /* send statistics event, offset > 0x40*/
            DsmApi_SendDeviceStatisticsEvent(longUpEvt->source, longUpEvt->subqualifier + 0x40, longUpEvt->data);
        } else if (longUpEvt->mainqualifier <= UpstreamMainqualifierParameter1) {
            /* send all parameters separated event */
            DsmApi_SendDeviceParameterEvent(longUpEvt->source, me->selectedParamBank, (longUpEvt->mainqualifier * 0x80 + longUpEvt->subqualifier * 2), (longUpEvt->data & 0xff));
            DsmApi_SendDeviceParameterEvent(longUpEvt->source, me->selectedParamBank, (longUpEvt->mainqualifier * 0x80 + longUpEvt->subqualifier * 2) + 1, (longUpEvt->data >> 8));
        } else if (longUpEvt->mainqualifier == UpstreamMainqualifierSensor) {
            if (longUpEvt->subqualifier == UpstreamSubqualifierSensorSubSub) {
                uint8_t subsubqualifier = longUpEvt->data & 0xFF;
                uint8_t data = (longUpEvt->data >> 8) & 0xFF;
                switch (subsubqualifier) {
                case UpstreamSubsubqualifierTestResponse: {
                    DsmApi_SendTestTransmissionQualityEvent(longUpEvt->source, data, longUpEvt->quality);
                    break;
                }
                case UpstreamSubsubqualifierUDIResponse: {
                    DsmApi_SendUDIResponseEvent(longUpEvt->source, data);
                    break;
                }
                }
            }
            else if (longUpEvt->subqualifier == UpstreamSubqualifierSensorSensorList) {
                uint16_t swapped_data = SWAP_BYTES(longUpEvt->data);
                uint8_t lastSensor = (swapped_data >> 15) & 0x1;
                uint8_t sensorPresent = (swapped_data >> 14) & 0x1;
                uint8_t sensorId = (swapped_data >> 8) & 0xF;
                uint8_t sensorType = swapped_data & 0xFF;
                DsmApi_SendSensorListEvent(longUpEvt->source, sensorId, sensorPresent, sensorType, lastSensor);
            }
            else if (longUpEvt->subqualifier == UpstreamSubqualifierSensorSensorValues) {
                uint16_t swapped_data = SWAP_BYTES(longUpEvt->data);
                uint8_t sensorId = (swapped_data >> 12) & 0xF;
                uint16_t sensorValue = (swapped_data) & 0xFFF;
                DsmApi_SendSensorValueEvent(longUpEvt->source, sensorId, sensorValue);
                {
                    DatamodelReqDeviceSensorPushEvt * reqEvt =
                            Q_NEW(DatamodelReqDeviceSensorPushEvt, DATAMODEL_REQ_DEVICE_PUSH_SENSOR_SIG);
                    reqEvt->deviceId = longUpEvt->source;
                    reqEvt->sensorIndex = sensorId;
                    reqEvt->sensorValue = sensorValue;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                }
            }
            else if (longUpEvt->subqualifier == UpstreamSubqualifierBinaryInputStates) {
                uint8_t binaryInputIndex = longUpEvt->index;
                uint8_t binaryInputState = longUpEvt->data;
                uint8_t binaryInputType = 0;
                if (Datamodel_GetDeviceBinaryInputType(me->deviceId, binaryInputIndex, &binaryInputType)) {
                    DsmApi_SendBinaryInputEvent(longUpEvt->source, binaryInputIndex, binaryInputType, binaryInputState);
                    vdsm_report(LOG_DEBUG, "selectB send binary input type. source:%i, binary index:%d, type:%d \n", longUpEvt->source, binaryInputIndex, binaryInputType);
                } else {
                    DsmApi_SendBinaryInputEvent(longUpEvt->source, binaryInputIndex, 0, binaryInputState);
                    vdsm_report(LOG_DEBUG, "selectB send binary input type. source:%i, binary index:%d\n", longUpEvt->source, binaryInputIndex);
                }
            }
            else if ((UpstreamSubqualifierSensorStatusStart <= longUpEvt->subqualifier) &&
                     (UpstreamSubqualifierSensorStatusEnd >= longUpEvt->subqualifier)) {
                uint16_t states = SWAP_BYTES(longUpEvt->data);
                DsmApi_SendSensorValueEvent(longUpEvt->source, longUpEvt->subqualifier, states);
            }
        }
        return Q_HANDLED();
    }
    case SELECT_B_CONFIGURE_PING_SIG: {
        ConfigurePingEvt const * configurePingEvt = (ConfigurePingEvt const *)e;
        switch (configurePingEvt->modifier) {
        case CIRCUIT_PING_DEVICES_AUTOMATIC:
            /* fall through */
        case CIRCUIT_PING_DEVICES_ONE_TIME_AUTOMATIC:
            /* fall through */
        case CIRCUIT_PING_DEVICES_ONE_TIME_OFF: {
            /* set interval times */
            if (configurePingEvt->oneTimeInterval == 0) {
                me->oneTimePingInterval = me->defaultOneTimePingInterval;
            } else if (configurePingEvt->oneTimeInterval < MIN_PING_INTERVAL_MS) {
                me->oneTimePingInterval = MIN_PING_INTERVAL_MS;
            } else {
                me->oneTimePingInterval = configurePingEvt->oneTimeInterval;
            }
            if (configurePingEvt->autoInterval == 0) {
                me->autoPingInterval = me->defaultAutoPingInterval;
            } else if (configurePingEvt->autoInterval < MIN_PING_INTERVAL_MS) {
                me->autoPingInterval = MIN_PING_INTERVAL_MS;
            } else {
                me->autoPingInterval = configurePingEvt->autoInterval;
            }
            if (configurePingEvt->modifier == CIRCUIT_PING_DEVICES_AUTOMATIC) {
                me->currentPingTicks = me->autoPingInterval / BSP_MS_PER_TICKS;
            } else {
                me->currentPingTicks = me->oneTimePingInterval / BSP_MS_PER_TICKS;
            }
            /* clear ping device ID */
            me->clearPingDeviceId = TRUE;
            /* reset ping counter */
            me->pingCount = 0;
            /* save new ping mode */
            me->pingMode = configurePingEvt->modifier;
            /* restart ping timer */
            QTimeEvt_disarm(&me->pingTimer);
            QTimeEvt_postIn(&me->pingTimer, (QActive *)me, me->currentPingTicks);
            break;
        }
        case CIRCUIT_PING_DEVICES_OFF: {
            me->pingDeviceId = 0;
            /* save new ping mode */
            me->pingMode = CIRCUIT_PING_DEVICES_OFF;
            /* stop ping timer */
            QTimeEvt_disarm(&me->pingTimer);
            break;
        }
        default: {
            return Q_HANDLED();
        }
        }
        if (me->returnToIdleState) {
            me->returnToIdleState = FALSE;
            return Q_TRAN(&SelectB_idle);
        }
        return Q_HANDLED();
    }
    case SELECT_B_PING_TIMEOUT_SIG: {
        /* set ping timer elapsed flag */
        me->pingTimerElapsed = TRUE;
        return Q_HANDLED();
    }
    case SELECT_B_DISABLE_JUMBLE_SIG: {
        DisableSelectBPartsEvt const * reqEvt = (DisableSelectBPartsEvt const *)e;
        me->jumbleDisabled = reqEvt->disable;
        return Q_HANDLED();
    }
    case SELECT_B_STOP_SIG: {
        return Q_TRAN(&SelectB_stopped);
    }
    case DOWNSTREAM_ERROR_SIG: {
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_MINUTE_TIMEOUT_SIG: {
        /* increment minute counter */
        ++me->minuteCount;
        /* send minute tick to heating controller */
        static QEvent const evt = { HEATING_MINUTE_TICK_SIG, 0 };
        QActive_postFIFO((QActive *)AO_SelectA, &evt);
        /* ping devices every 24 hours */
        if ((me->minuteCount % SELECT_B_MINUTES_PER_24_HOURS) == 0) {
            if (me->pingEvery24HoursEnabled && (me->pingDeviceId == 0)) {
                /* 24h ping is enabled and no ping round is in progress */
                static QEvent const evt = { SELECT_B_PING_START_24_HOURS_PING_ROUND_SIG, 0 };
                QActive_postFIFO((QActive *)me, &evt);
            }
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState SelectB_idle(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* recall deferred command event? */
        if (QActive_recall((QActive *)me, &me->deferredCommandQueue) != 0) {
            return Q_HANDLED();
        }
        /* start ping interval timer */
        if (me->pingTimerElapsed) {
            static QEvent const pingTimeoutEvt = { SELECT_B_PING_TIMEOUT_SIG, 0 };
            QActive_postLIFO((QActive *)me, &pingTimeoutEvt);
        } else {
            /* recall inactive device to ping (lowest priority) */
            QActive_recall((QActive *)me, &me->deferredInactiveDevToPingQueue);
        }
        return Q_HANDLED();
    }
    case SELECT_B_PING_START_24_HOURS_PING_ROUND_SIG: {
        /* set ping mode to one-time off */
        me->pingMode = CIRCUIT_PING_DEVICES_ONE_TIME_OFF;
        /* clear ping device ID */
        me->clearPingDeviceId = TRUE;
        /* a ping round might already ongoing */
        QTimeEvt_disarm(&me->pingTimer);
        /* fall through */
    }
    case SELECT_B_PING_TIMEOUT_SIG: {
        /* clear ping timer elapsed flag */
        me->pingTimerElapsed = FALSE;
        /* is ping mode off? */
        if (me->pingMode == CIRCUIT_PING_DEVICES_OFF) {
            /* just exit */
            return Q_TRAN(&SelectB_idle);
        }
        if (me->clearPingDeviceId) {
            me->clearPingDeviceId = FALSE;
            me->pingDeviceId = 0;
        }
        if (me->pingDeviceId == 0) {
            /* get first device to ping */
            return Q_TRAN(&SelectB_ping_get_device_info);
        }
        /* ping device */
        return Q_TRAN(&SelectB_ping_send_ping);
    }
    case SELECT_B_GET_DEVICE_SENSOR_SIG: {
        GetDeviceSensorEvt const * reqEvt = (GetDeviceSensorEvt const *)e;
        vdc_api_device_get_sensor_value(reqEvt->deviceId, reqEvt->sensorId);
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_GET_DEVICE_STATISTICS_SIG: {
        GetDeviceStatisticsEvt const * reqEvt = (GetDeviceStatisticsEvt const *)e;
        /* save request parameters */
        me->getStatisticsDeviceId = reqEvt->deviceId;
        me->getStatisticsParamOffset = reqEvt->paramOffset;
        /* TODO: ?? get device statistics */
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_GET_DEVICE_PARAMETER_SIG: {
        char dsuid[DSUID_STR_LEN];
        DeviceParameterEvt const * reqEvt = (DeviceParameterEvt const *)e;
        me->selectedParamBank = reqEvt->paramBank;
        if (Datamodel_GetDsuidFromDeviceId(reqEvt->deviceId, dsuid)) {
            device_api_get_parameter(dsuid, reqEvt->paramBank, reqEvt->paramOffset);
        } else {
            vdsm_report(LOG_WARNING, "selectB get device parameter. device: %d not found\n", reqEvt->deviceId);
        }
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_OUTPUT_CHANNEL_SIG: {

        DeviceOutpuChannelEvt const * reqEvt = (DeviceOutpuChannelEvt const *)e;
        /* save request parameters */
        me->deviceId = reqEvt->deviceId;
        me->OPCchannel = reqEvt->channel;
        me->OPCoperation = reqEvt->operation;
        me->OPCvalue = reqEvt->value;
        me->OPCscene = reqEvt->scene;
        me->OPCis16bit = reqEvt->is16bit;

        /* set get device parameter history */
        me->returnState = (QStateHandler)&SelectB_idle;
        me->superState = (QStateHandler)&SelectB_busy;

        switch (me->OPCoperation) {
        case DEVICE_O_P_C_CONFIG_SET_CURRENT:
        case DEVICE_O_P_C_CONFIG_SET_CURRENT_AND_APPLY: {
            bool apply = (me->OPCoperation == DEVICE_O_P_C_CONFIG_SET_CURRENT_AND_APPLY);
            vdc_api_device_set_opc_output_state(me->deviceId, me->OPCchannel,
                                                me->OPCvalue, apply);
            return Q_TRAN(&SelectB_idle);
        }
        break;

        case DEVICE_O_P_C_CONFIG_GET_CURRENT:
            // get output state
            me->selectedParamBank = 100;
            vdc_api_device_get_opc_output_state(me->deviceId, me->OPCchannel);
            return Q_HANDLED();
        break;

        case DEVICE_O_P_C_CONFIG_SET_SCENE:
            vdc_api_device_set_opc_scene(me->deviceId, me->OPCscene,
                                         me->OPCchannel, me->OPCvalue);
            return Q_TRAN(&SelectB_idle);
        break;

        case DEVICE_O_P_C_CONFIG_SET_SCENE_CONFIG:
            vdc_api_device_set_opc_scene_config(me->deviceId, me->OPCscene, me->OPCvalue);
            return Q_TRAN(&SelectB_idle);

        case DEVICE_O_P_C_CONFIG_SET_DC_FLAGS:
            vdc_api_device_set_opc_scene_dc_flags(me->deviceId, me->OPCscene, me->OPCvalue);
            return Q_TRAN(&SelectB_idle);
        break;

        case DEVICE_O_P_C_CONFIG_GET_SCENE:
            me->selectedParamBank = 102;
            vdc_api_device_get_opc_scene_value(me->deviceId, me->OPCscene,
                                               me->OPCchannel);
            return Q_HANDLED();
        break;

        case DEVICE_O_P_C_CONFIG_GET_SCENE_CONFIG:
            me->selectedParamBank = 103;
            vdc_api_device_get_opc_scene_config(me->deviceId, me->OPCscene,
                                                me->OPCoperation);
            return Q_TRAN(&SelectB_idle);

        case DEVICE_O_P_C_CONFIG_GET_DC_FLAGS:
            me->selectedParamBank = 104;
            vdc_api_device_get_opc_scene_config(me->deviceId, me->OPCscene,
                                                me->OPCoperation);
            return Q_TRAN(&SelectB_idle);

        default:
            /* never come here */
            vdsm_report(LOG_ERR, "Switch not supported:%d\n", me->OPCoperation);
            Q_ASSERT(FALSE);
        }
    }
    case SELECT_B_SET_DEVICE_PARAMETER_SIG: {
        DeviceParameterEvt const * reqEvt = (DeviceParameterEvt const *)e;
        uint8_t error = ERROR_OK;
        /* send modification to vdc device */
        if (vdc_api_device_set_register(reqEvt->deviceId, reqEvt->paramBank, reqEvt->paramOffset, reqEvt->value)) {
            if (reqEvt->paramBank == 3) {
                switch (reqEvt->paramOffset) {
                case MODE: {
                    /* output mode was modified, adjust internal output_load flag */
                    DatamodelReqDeviceSetMODERegEvt * evt =
                        Q_NEW(DatamodelReqDeviceSetMODERegEvt, DATAMODEL_REQ_DEVICE_SET_OUTPUT_LOAD_SIG);
                    evt->basic.requester = (QActive *)me;
                    evt->deviceAddress = reqEvt->deviceId;
                    evt->MODERegister = reqEvt->value;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)evt);
                    break;
                }
                case BUTTON_ID_GRP_0: {
                    /* Button0 Id or Button0 Group was modified, adjust Datamodel */
                    DatamodelReqDeviceSetButtonIdAndGroupRegEvt * evt =
                        Q_NEW(DatamodelReqDeviceSetButtonIdAndGroupRegEvt, DATAMODEL_REQ_DEVICE_SET_BUTTON_ID_AND_GROUP_SIG);
                    evt->basic.requester = (QActive *)me;
                    evt->deviceAddress = reqEvt->deviceId;
                    evt->buttonIdAndGroupRegister = reqEvt->value;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)evt);
                    break;
                }
                case LTMODE: {
                    /* ltmode register was modified, update datamodel */
                    DatamodelReqDeviceSetLTMODERegEvt * evt =
                        Q_NEW(DatamodelReqDeviceSetLTMODERegEvt, DATAMODEL_REQ_DEVICE_SET_LTMODE_SIG);
                    evt->basic.requester = (QActive *)me;
                    evt->deviceAddress = reqEvt->deviceId;
                    evt->LTMODERegister = reqEvt->value;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)evt);
                    break;
                }
                }
            }
        } else {
            error = ERROR_DEVICE_SELECT_RESPONSE_TIMEOUT;
        }
        /* send done event */
        SelectB_SendDsmDoneEvent(reqEvt, error);
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_USER_PROTOBUF_MESSAGE_REQUEST_SIG: {
        SelectBProtobufMessageRequestEvt const * reqEvt =
            (SelectBProtobufMessageRequestEvt const *)e;
        vdc_api_user_protobuf_message_request(reqEvt->size, reqEvt->request, &reqEvt->dsuid);
        return Q_TRAN(&SelectB_user_protobuf_message_request);
    }
    case SELECT_B_TEST_TRANSMISSION_QUALITY_SIG: {
        GetTransmissionQualityEvt const * reqEvt = (GetTransmissionQualityEvt const *)e;
        DsmApi_SendTestTransmissionQualityEvent(reqEvt->deviceId, 0, 62);
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_GET_DEVICE_DSLINK_SIG: {
        GetDevicedSLinkEvt const * reqEvt = (GetDevicedSLinkEvt const *)e;
        /* save device */
        me->deviceId = reqEvt->deviceId;
        me->dSLinkInfo.operation = reqEvt->operation;
        me->dSLinkInfo.value = reqEvt->value;
        /* TODO: ?? */
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_GET_ZONE_GROUP_DSLINK_SIG: {
        GetZoneGroupdSLinkEvt const * reqEvt = (GetZoneGroupdSLinkEvt const *)e;
        /* save device */
        me->localZoneId = reqEvt->localZoneId;
        me->groupId = reqEvt->groupId;
        me->dSLinkInfo.operation = reqEvt->operation;
        me->dSLinkInfo.value = reqEvt->value;
        /* TODO: ?? */
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_SET_ZONE_GROUP_EXTRA_COMMAND: {
        SetZoneGroupExtraCommandEvt const * reqEvt = (SetZoneGroupExtraCommandEvt const *)e;
        me->localZoneId = reqEvt->localZoneId;
        me->groupId = reqEvt->groupId;
        me->extraCommand = reqEvt->command;
        me->extraCommandDeviceSelect = FALSE;
        /* TODO ?? */
        // set locked unlocked to vdcd is not supported
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_PING_INACTIVE_DEVICE_SIG: {
        me->pingSingleDeviceId = ((PingInactiveDeviceEvt const *)e)->deviceId;
        me->pingSingleTriggeredByInvalid = TRUE;
        me->returnState = (QStateHandler)&SelectB_idle;
        me->superState = (QStateHandler)&SelectB_busy;
        me->pingSingleDeviceCount = 0;
        return Q_TRAN(&SelectB_ping_single_device);
    }
    case SELECT_B_TEST_DEVICE_PRESENCE_SIG: {
        me->pingSingleDeviceId = ((GetDevicePresenceEvt const *)e)->deviceId;
        me->pingSingleTriggeredByInvalid = FALSE;
        me->returnState = (QStateHandler)&SelectB_idle;
        me->superState = (QStateHandler)&SelectB_busy;
        me->pingSingleDeviceCount = 0;
        return Q_TRAN(&SelectB_ping_single_device);
    }
    case SENSOR_PUSH_VALUE_SIG: {
        ZoneGroupSensorPushEvt const * reqEvt = (ZoneGroupSensorPushEvt const *)e;
        if (reqEvt->localZoneId != DATAMODEL_LOCAL_ZONE_ID_NOT_APPLICABLE) {
            /* sensor push message is addressed to local zone */
            vdc_api_zone_push_sensor_value(reqEvt->localZoneId, reqEvt->groupId, reqEvt->sensorType, reqEvt->sensorValue);
        }
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_ANNOUNCE_DEVICE_SIG: {
        SelectBAnnounceDeviceEvt const * evt = (SelectBAnnounceDeviceEvt const *)e;
        strncpy(me->dsuid, evt->dsuid, sizeof(me->dsuid));
        strncpy(me->vdc_dsuid, evt->vdcdsuid, sizeof(me->vdc_dsuid));
        me->messageId = evt->message_id;
        return Q_TRAN(SelectB_register_lookup_device);
    }
    }
    return Q_SUPER(&SelectB_top);
}

static QState SelectB_register_lookup_device(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
    {
        DatamodelReqDeviceSerialEvt * reqEvt = Q_NEW(DatamodelReqDeviceSerialEvt, DATAMODEL_REQ_DEVICE_ENTRY_BY_SERIAL_NUMBER_SIG);
        reqEvt->basic.requester = (QActive *)me;
        strncpy(reqEvt->dsuid, me->dsuid, sizeof(reqEvt->dsuid));
        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
        return Q_HANDLED();
    }
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        DatamodelResDeviceEntryEvt const * resEvt = (DatamodelResDeviceEntryEvt const *)e;
        /* save device properties */
        me->deviceId = resEvt->device.deviceId;
        me->vid = resEvt->device.vid;
        me->localZoneId = resEvt->device.localZoneId;
        me->deviceIsKnown = TRUE;

        device_api_get_parameter(me->dsuid, 1, FID);
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, SELECT_B_WAIT_FOR_DSID_TICKS);
        return Q_HANDLED();
    }
    case SELECT_B_FILTERED_LONG_UPSTREAM_SIG: {
        LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e;
        if ((longUpEvt->mainqualifier == 0) && (longUpEvt->subqualifier == (FID / 2))) {
            QTimeEvt_disarm(&me->timeEvt);
            me->fid = longUpEvt->data;
            Datamodel_updateDeviceVdc(me->deviceId, me->vdc_dsuid, me->fid);
            if (Fid_HasBinaryInputs(me->fid)) {
                return Q_TRAN(&SelectB_register_get_sensors);
            } else {
                return Q_TRAN(&SelectB_register_get_LTMode_param);
            }
        }
        break;
    }
    case DATAMODEL_RES_ERROR_SIG: {
        DatamodelLongErrorEvt const * errorEvt = (DatamodelLongErrorEvt const *)e;
        if (errorEvt->error == ERROR_DEVICE_NOT_FOUND) {
            /* device was not found -> get params of device */
            me->deviceIsKnown = FALSE;
            // unknown at this point. Device not created in database yet.
            me->deviceId = 0; 
            return Q_TRAN(&SelectB_register_get_zone);
        }
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_TIMEOUT_SIG: {
        /* exit registration */
        vdsm_report(LOG_WARNING, "TIMEOUT SelectB_register_lookup_device\n");
        return Q_TRAN(&SelectB_idle);
    }
    }
    return Q_SUPER(&SelectB_busy);
}

/*..........................................................................*/
static QState SelectB_register_get_zone(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        vdc_api_device_get_zone(me->dsuid);

        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, SELECT_B_WAIT_FOR_DSID_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case SELECT_B_TIMEOUT_SIG: {
        vdsm_report(LOG_WARNING, "TIMEOUT SelectB_register_get_zone\n");
        me->zoneId = 0; // the zone check is implemented in datamodel
                        // see Datamodel_InitDevice

        return Q_TRAN(&SelectB_register_get_params);
    }
    case SELECT_B_FILTERED_LONG_UPSTREAM_SIG: {
        LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e;
        me->zoneId = longUpEvt->data;
        return Q_TRAN(&SelectB_register_get_params);
        break;
    }
    }
    return Q_SUPER(&SelectB_busy);
}

/*..........................................................................*/
static QState SelectB_register_get_params(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* send READ_PARA16_B */
        device_api_get_parameter(me->dsuid, 1, FID);
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, SELECT_B_WAIT_FOR_DSID_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case SELECT_B_TIMEOUT_SIG: {
        /* exit registration */
        vdsm_report(LOG_WARNING, "TIMEOUT SelectB_register_get_params\n");
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_FILTERED_LONG_UPSTREAM_SIG: {
        LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e;
        if (longUpEvt->mainqualifier == 0) {
            uint16_t data = longUpEvt->data;
            me->fid = data;

            /* we have all parameters, now add device */
            DatamodelReqAddDeviceEvt * reqEvt =
                Q_NEW(DatamodelReqAddDeviceEvt, DATAMODEL_REQ_ADD_DEVICE_SIG);
            reqEvt->basic.requester = (QActive *)me;
            memcpy(reqEvt->dsuid, me->dsuid, sizeof(reqEvt->dsuid));
            memcpy(reqEvt->vdcdsuid, me->vdc_dsuid, sizeof(reqEvt->vdcdsuid));

            reqEvt->vid = 0xffff;
            reqEvt->fid = me->fid;
            reqEvt->ver = 0xffff;
            reqEvt->zoneId = me->zoneId; // the zone check is implemented in datamodel
            QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
            return Q_HANDLED();
        }
        break;
    }
    case DATAMODEL_RES_ADD_DEVICE_SIG: {
        DatamodelResAddDeviceEvt const * resEvt = (DatamodelResAddDeviceEvt const *)e;
        /* save device ID */
        me->deviceId = resEvt->deviceId;
        me->localZoneId = resEvt->localZoneId;

        /* set zone on device */
        if (!device_api_set_zone(me->dsuid, resEvt->zoneId)) {
            vdsm_report(LOG_ERR, "selectB could not set zone to device: %d.\n", me->deviceId);
        }

        if (Fid_HasBinaryInputs(me->fid)) {
            return Q_TRAN(&SelectB_register_get_sensors);
        } else {
            return Q_TRAN(&SelectB_register_get_LTMode_param);
        }
    }
    case ERROR_REQUEST_CAN_NOT_BE_EXECUTED: {
        return Q_TRAN(&SelectB_idle);
    }
    }
    return Q_SUPER(&SelectB_busy);
}

/*..........................................................................*/
static QState SelectB_register_get_sensors(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        vdc_api_device_register_sensors(me->deviceId);
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, SELECT_B_WAIT_FOR_DSID_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case SELECT_B_TIMEOUT_SIG: {
        /* exit registration */
        vdsm_report(LOG_WARNING, "TIMEOUT SelectB_register_get_sensors\n");
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_FILTERED_LONG_UPSTREAM_SIG: {
        LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e;
        if (longUpEvt->source == me->deviceId) {
            return Q_TRAN(&SelectB_register_get_binary_inputs);
        }
        break;
    }
    }
    return Q_SUPER(&SelectB_busy);
}

/*..........................................................................*/
static QState SelectB_register_get_binary_inputs(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        vdc_api_device_register_binary_inputs(me->deviceId);
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, SELECT_B_WAIT_FOR_DSID_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case SELECT_B_TIMEOUT_SIG: {
        /* exit registration */
        vdsm_report(LOG_WARNING, "TIMEOUT SelectB_register_get_binary_inputs\n");
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_FILTERED_LONG_UPSTREAM_SIG: {
        LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e;
        if (longUpEvt->source == me->deviceId) {
            return Q_TRAN(&SelectB_register_get_LTMode_param);
        }
        break;
    }
    }
    return Q_SUPER(&SelectB_busy);
}

/*..........................................................................*/
static QState SelectB_register_get_LTMode_param(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* send READ_PARA16_B */
        char dsuid[DSUID_STR_LEN];
        if (Datamodel_GetDsuidFromDeviceId(me->deviceId, dsuid)) {
            device_api_get_parameter(dsuid,  3, LTMODE);
        } else {
            vdsm_report(LOG_WARNING, "selectB get device parameter ltmode. device: %d not found\n", me->deviceId);
        }
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, SELECT_B_WAIT_FOR_DSID_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case SELECT_B_TIMEOUT_SIG: {
        /* exit registration */
        vdsm_report(LOG_WARNING, "TIMEOUT SelectB_register_get_LTMode_param\n");
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_FILTERED_LONG_UPSTREAM_SIG: {
        LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e;
        if ((longUpEvt->mainqualifier == (LTMODE / 0x80)) && (longUpEvt->subqualifier == (LTMODE / 2))) {
            {
                DatamodelReqDeviceSetLTMODERegEvt * reqEvt =
                    Q_NEW(DatamodelReqDeviceSetLTMODERegEvt, DATAMODEL_REQ_DEVICE_SET_LTMODE_SIG);
                reqEvt->basic.requester = (QActive *)me;
                reqEvt->deviceAddress = me->deviceId;
                reqEvt->LTMODERegister = (longUpEvt->data & 0xFF);
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
            }
            if (Fid_HasVariableButtonIdGroup(me->fid) || Fid_HasOutput(me->fid)) {
                return Q_TRAN(&SelectB_register_get_Mode_param);
            }
        return Q_TRAN(&SelectB_idle);
        }
        break;
    }
    }
    return Q_SUPER(&SelectB_busy);
}

/*..........................................................................*/
static QState SelectB_register_get_Mode_param(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* send READ_PARA16_B */
        char dsuid[DSUID_STR_LEN];
        if (Datamodel_GetDsuidFromDeviceId(me->deviceId, dsuid)) {
            device_api_get_parameter(dsuid,  3, MODE);
        } else {
            vdsm_report(LOG_WARNING, "selectB get device parameter mode. device: %d not found\n", me->deviceId);
        }
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, SELECT_B_WAIT_FOR_DSID_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case SELECT_B_TIMEOUT_SIG: {
        /* exit registration */
        vdsm_report(LOG_WARNING, "TIMEOUT SelectB_register_get_Mode_param\n");
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_FILTERED_LONG_UPSTREAM_SIG: {
        LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e;
        if ((longUpEvt->mainqualifier == (MODE / 0x80)) && (longUpEvt->subqualifier == (MODE / 2))) {
            uint16_t data = longUpEvt->data;
            if (Fid_HasOutput(me->fid)) {
                DatamodelReqDeviceSetMODERegEvt * reqEvt =
                    Q_NEW(DatamodelReqDeviceSetMODERegEvt, DATAMODEL_REQ_DEVICE_SET_OUTPUT_LOAD_SIG);
                reqEvt->basic.requester = (QActive *)me;
                reqEvt->deviceAddress = me->deviceId;
                reqEvt->MODERegister = (data & 0xFF);
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
            }
            /* read button ID and group registers? */
            if (Fid_HasVariableButtonIdGroup(me->fid) && Fid_HasButton(me->fid)) {
                DatamodelReqDeviceSetButtonIdAndGroupRegEvt * reqEvt =
                    Q_NEW(DatamodelReqDeviceSetButtonIdAndGroupRegEvt, DATAMODEL_REQ_DEVICE_SET_BUTTON_ID_AND_GROUP_SIG);
                reqEvt->basic.requester = (QActive *)me;
                reqEvt->deviceAddress = me->deviceId;
                reqEvt->buttonIdAndGroupRegister = ((data >> 8) & 0xFF);
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
            }
            return Q_TRAN(&SelectB_register_get_group_membership);
        }
        break;
    }
    }
    return Q_SUPER(&SelectB_busy);
}

/*..........................................................................*/
static QState SelectB_register_get_group_membership(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        vdc_api_device_get_group_membership(me->deviceId);
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, SELECT_B_WAIT_FOR_DSID_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case SELECT_B_TIMEOUT_SIG: {
        /* exit registration */
        vdsm_report(LOG_WARNING, "TIMEOUT SelectB_register_get_group_membership\n");
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_FILTERED_LONG_UPSTREAM_SIG: {
        LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e;
        if (longUpEvt->source == me->deviceId) {
            return Q_TRAN(&SelectB_register_get_Opc_param);
        }
        break;
    }
    }
    return Q_SUPER(&SelectB_busy);
}

/*..........................................................................*/
static QState SelectB_register_get_Opc_param(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        vdc_api_device_register_opc_channels(me->deviceId);

        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, SELECT_B_WAIT_FOR_DSID_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case SELECT_B_TIMEOUT_SIG: {
        /* exit registration */
        vdsm_report(LOG_WARNING, "TIMEOUT SelectB_register_get_Opc_param\n");
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_FILTERED_LONG_UPSTREAM_SIG: {
        LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e;
        if (longUpEvt->source == me->deviceId) {
            return Q_TRAN(&SelectB_register_get_name);
        }
        break;
    }
    }
    return Q_SUPER(&SelectB_busy);
}

/*..........................................................................*/
static QState SelectB_register_get_name(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        vdc_api_device_get_name(me->deviceId);
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, SELECT_B_WAIT_FOR_DSID_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case SELECT_B_TIMEOUT_SIG: {
        /* exit registration */
        vdsm_report(LOG_WARNING, "TIMEOUT SelectB_register_get_name\n");
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_FILTERED_LONG_UPSTREAM_SIG: {
        LongUpstreamEvt const * longUpEvt = (LongUpstreamEvt const *)e;
        if (longUpEvt->source == me->deviceId) {
            DatamodelZoneInfo * zoneInfo = Datamodel_GetZoneInfoByLocalId(me->localZoneId);
            Q_ASSERT(zoneInfo != 0);
            dsuid_t dsuid;
            dsuid_from_string(me->dsuid, &dsuid);
            if (me->deviceIsKnown) {
                /* send device accessibility on event for AKM */
                if (Fid_HasBinaryInputs(me->fid)) {
                    DsmApi_SendDeviceAccessibilityEvent(me->deviceId, zoneInfo->super.zoneId, me->vid, 0 /*no pid for non ds devices*/, me->fid, me->ver, dsuid, TRUE);
                }
            } else {
                /* send device accessibility */
                DsmApi_SendDeviceAccessibilityEvent(me->deviceId, zoneInfo->super.zoneId, me->vid, 0 /*no pid for non ds devices*/, me->fid, me->ver, dsuid, TRUE);
            }
            /* send device changed event */
            DsmApi_SendDeviceModelChangedEvent(me->deviceId);
            vdc_api_device_send_announce_ok(me->messageId, me->deviceId);
            return Q_TRAN(&SelectB_idle);
        }
        break;
    }
    }
    return Q_SUPER(&SelectB_busy);
}

/*..........................................................................*/
static QState SelectB_user_protobuf_message_request(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, SELECT_B_WAIT_FOR_DSID_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case SELECT_B_TIMEOUT_SIG: {
        static const QEvent reqEvt = { SELECT_B_USER_PROTOBUF_MESSAGE_TIMEOUT_SIG, 0 };
        QActive_postFIFO(AO_DsmApi, &reqEvt);
        return Q_TRAN(&SelectB_idle);
    }
    case VDC_API_USER_PROTOBUF_MESSAGE_RESPONSE_SIG: {
        VdcUserProtobufMessageResponseEvt const *resEvt =
            (VdcUserProtobufMessageResponseEvt const *)e;
        SelectBProtobufMessageResponseEvt *reqEvt =
            Q_NEW(SelectBProtobufMessageResponseEvt, SELECT_B_USER_PROTOBUF_MESSAGE_RESPONSE_SIG);

        reqEvt->size = resEvt->size;
        memcpy(reqEvt->response, resEvt->response, reqEvt->size);
        QActive_postFIFO(AO_DsmApi, (QEvent *)reqEvt);
        return Q_TRAN(&SelectB_idle);
    }
    }
    return Q_SUPER(&SelectB_busy);
}

/*..........................................................................*/
static QState SelectB_busy(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case SELECT_B_SET_DSM_ID_SIG: {
        if (me->deviceAddress != DS485_INVALID_ADDRESS) {
            me->setDsmId = TRUE;
        }
        return Q_HANDLED();
    }
    case SELECT_B_HELLO_SIG: {
        me->helloReceived = TRUE;
        return Q_HANDLED();
    }
    case SELECT_B_SET_ZONE_GROUP_EXTRA_COMMAND:
        /* fall through */
    case SELECT_B_USER_PROTOBUF_MESSAGE_REQUEST_SIG:
        /* fall through */
    case SELECT_B_ANNOUNCE_DEVICE_SIG:
        /* fall through */
    case DATAMODEL_REQ_ADD_DEVICE_SIG:
        /* fall through */
    case SELECT_B_SELECT_SIG:
        /* fall through */
    case SELECT_B_PING_START_24_HOURS_PING_ROUND_SIG:
        /* fall through */
    case SELECT_B_TEST_DEVICE_PRESENCE_SIG:
        /* fall through */
    case SENSOR_PUSH_VALUE_SIG:
        /* fall through */
    case SELECT_B_GET_ZONE_GROUP_DSLINK_SIG:
        /* fall through */
    case SELECT_B_GET_DEVICE_DSLINK_SIG:
        /* fall through */
    case SELECT_B_GET_DEVICE_SENSOR_SIG:
        /* fall through */
    case SELECT_B_RESTART_REGISTRATION:
        /* fall through */
    case SELECT_B_GET_DEVICE_PARAMETER_SIG:
        /* fall through */
    case SELECT_B_SET_DEVICE_PARAMETER_SIG:
        /* fall through */
    case SELECT_B_OUTPUT_CHANNEL_SIG:
        /* fall through */
    case SELECT_B_GET_DEVICE_STATISTICS_SIG:
        /* fall through */
    case SELECT_B_TEST_TRANSMISSION_QUALITY_SIG: {
        if (QEQueue_getNFree(&me->deferredCommandQueue) != 0) {
            /* defer command */
            QActive_defer((QActive *)me, &me->deferredCommandQueue, e);
        } else {
            vdsm_report(LOG_ERR, "selectB throwing away message\n");
        }
        return Q_HANDLED();
    }
    case SELECT_B_FILTERED_LONG_UPSTREAM_SIG: {
        return Q_HANDLED();
    }
    case SELECT_B_PING_INACTIVE_DEVICE_SIG: {
        if (QEQueue_getNFree(&me->deferredInactiveDevToPingQueue) != 0) {
            /* defer inactive device to ping event */
            QActive_defer((QActive *)me, &me->deferredInactiveDevToPingQueue, e);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&SelectB_top);
}

/*..........................................................................*/
static QState SelectB_ping_busy(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case SELECT_B_CONFIGURE_PING_SIG: {
        /* some ping configuration were received -> return to idle state after reconfiguration */
        me->returnToIdleState = TRUE;
        break;
    }
    case DOWNSTREAM_ERROR_SIG: {
        /* reset ping counter */
        me->pingCount = 0;
        /* restart ping timer */
        QTimeEvt_postIn(&me->pingTimer, (QActive *)me, me->currentPingTicks);
        break;
    }
    case DATAMODEL_RES_ERROR_SIG: {
        /* no device found */
        /* clear ping device ID */
        me->clearPingDeviceId = TRUE;
        /* reset ping counter */
        me->pingCount = 0;
        /* restart ping timer */
        QTimeEvt_postIn(&me->pingTimer, (QActive *)me, SELECT_B_GET_FIRST_PING_DEVICE_TICKS);
        return Q_TRAN(&SelectB_idle);
    }
    }
    return Q_SUPER(&SelectB_busy);
}

/*..........................................................................*/
static QState SelectB_ping_get_device_info(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DatamodelReqDeviceInfoOfIndexEvt * reqEvt =
            Q_NEW(DatamodelReqDeviceInfoOfIndexEvt, DATAMODEL_REQ_DEVICE_INFO_OF_INDEX_SIG);
        reqEvt->basic.requester = (QActive *)me;
        reqEvt->zoneIdModifier = ZONE_INFO_BY_INDEX;
        reqEvt->modifier = DEVICE_INFO_BY_INDEX;
        reqEvt->zoneId = 0;
        reqEvt->deviceIndex = 0;
        reqEvt->privateId = 0;

        /* get vdc dsuid form last device */
        dsuid_t deviceDsuidId;
        dsuid_t vdcDsuidId;
        dsuid_from_string(me->dsuid, &deviceDsuidId);
        Datamodel_GetVdcDsuidFromDeviceDsuid(&deviceDsuidId, &vdcDsuidId);

        reqEvt->dsuid = vdcDsuidId;
        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
        return Q_HANDLED();
    }
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        /* first device found */
        DatamodelResDeviceEntryEvt * const resEvt = (DatamodelResDeviceEntryEvt * const)e;
        me->pingDeviceId = resEvt->device.deviceId;
        /* set one time ping initial device ID */
        me->oneTimePingInitialDeviceId = me->pingDeviceId;
        /* ping device */
        return Q_TRAN(&SelectB_ping_send_ping);
    }
    }
    return Q_SUPER(&SelectB_ping_busy);
}

/*..........................................................................*/
static QState SelectB_ping_send_ping(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        vdc_api_device_ping(me->pingDeviceId);
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, SELECT_B_WAIT_FOR_DSID_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&me->txTimer);
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case SELECT_B_TIMEOUT_SIG: {
        /* increment ping counter */
        if (++me->pingCount > 2) {
            /* device is inactive */
            DatamodelReqPingDeviceStateEvt * reqEvt =
                Q_NEW(DatamodelReqPingDeviceStateEvt, DATAMODEL_REQ_PING_DEVICE_STATE_SIG);
            reqEvt->basic.requester = (QActive *)me;
            reqEvt->deviceId = me->pingDeviceId;
            reqEvt->oneTimePingInitialDeviceId = me->oneTimePingInitialDeviceId;
            reqEvt->active = FALSE;
            QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
            return Q_TRAN(&SelectB_ping_get_next_device);
        }
        /* restart ping timer */
        QTimeEvt_postIn(&me->pingTimer, (QActive *)me, me->currentPingTicks);
        return Q_TRAN(&SelectB_idle);
    }
    case SELECT_B_PING_RESPONSE_SIG: {
        PingResponseEvt const * respEvt = (PingResponseEvt const *)e;
        /* check dSID */
        if (respEvt->deviceId == me->pingDeviceId) {
            DatamodelReqPingDeviceStateEvt * reqEvt;
            /* device is active */
            reqEvt = Q_NEW(DatamodelReqPingDeviceStateEvt, DATAMODEL_REQ_PING_DEVICE_STATE_SIG);
            reqEvt->basic.requester = (QActive *)me;
            reqEvt->deviceId = me->pingDeviceId;
            reqEvt->oneTimePingInitialDeviceId = me->oneTimePingInitialDeviceId;
            reqEvt->active = TRUE;
            QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
            return Q_TRAN(&SelectB_ping_get_next_device);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&SelectB_ping_busy);
}

/*..........................................................................*/
static QState SelectB_ping_get_next_device(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_PING_DEVICE_SIG: {
        DatamodelResPingDeviceEvt const * resEvt = (DatamodelResPingDeviceEvt const *)e;
        /* save next device ID to ping */
        me->pingDeviceId = resEvt->nextDeviceId;
        /* reset ping counter */
        me->pingCount = 0;
        /* stop one-time ping round? */
        if (resEvt->oneTimePingDone) {
            /* reset one-time ping initial device ID */
            me->oneTimePingInitialDeviceId = 0;
            /* change ping mode? */
            if (me->pingMode == CIRCUIT_PING_DEVICES_ONE_TIME_OFF) {
                /* don't restart ping timer */
                /* change to off mode */
                me->pingMode = CIRCUIT_PING_DEVICES_OFF;
                /* reset device ID to ping */
                me->pingDeviceId = 0;
            } else {
                if (me->pingMode == CIRCUIT_PING_DEVICES_ONE_TIME_AUTOMATIC) {
                    /* change current ping interval */
                    me->currentPingTicks = me->autoPingInterval / BSP_MS_PER_TICKS;
                    /* change to auto mode */
                    me->pingMode = CIRCUIT_PING_DEVICES_AUTOMATIC;
                }
                /* restart timer */
                QTimeEvt_postIn(&me->pingTimer, (QActive *)me, me->currentPingTicks);
            }
        } else {
            /* just restart timer */
            QTimeEvt_postIn(&me->pingTimer, (QActive *)me, me->currentPingTicks);
        }
        return Q_TRAN(&SelectB_idle);
    }
    }
    return Q_SUPER(&SelectB_ping_busy);
}

/*..........................................................................*/
static QState SelectB_ping_single_device(SelectB *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        vdc_api_device_ping(me->pingSingleDeviceId);
        QTimeEvt_postIn(&me->timeEvt, (QActive *)me, SELECT_B_WAIT_FOR_DSID_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&me->txTimer);
        QTimeEvt_disarm(&me->timeEvt);
        return Q_HANDLED();
    }
    case SELECT_B_TIMEOUT_SIG: {
        /* increment ping counter */
        if (++me->pingSingleDeviceCount > 2) {
            /* device is inactive */
            if (!me->pingSingleTriggeredByInvalid) {
                DsmApi_SendDevicePresenceEvent(me->pingSingleDeviceId, FALSE);
            }
            {
                DatamodelReqPingDeviceStateEvt * reqEvt;
                reqEvt = Q_NEW(DatamodelReqPingDeviceStateEvt, DATAMODEL_REQ_PING_DEVICE_STATE_SIG);
                reqEvt->basic.requester = 0;
                reqEvt->deviceId = me->pingSingleDeviceId;
                reqEvt->active = FALSE;
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
            }
            return Q_TRAN(me->returnState);
        }
        /* retry */
        me->selectTopState = me->superState;
        return Q_TRAN(&SelectB_ping_single_device);
    }
    case SELECT_B_PING_RESPONSE_SIG: {
        PingResponseEvt const * respEvt = (PingResponseEvt const *)e;
        /* check dSID */
        if (respEvt->deviceId == me->pingSingleDeviceId) {
            /* device is active */
            if (!me->pingSingleTriggeredByInvalid) {
                DsmApi_SendDevicePresenceEvent(me->pingSingleDeviceId, TRUE);
            }
            {
                DatamodelReqPingDeviceStateEvt * reqEvt;
                reqEvt = Q_NEW(DatamodelReqPingDeviceStateEvt, DATAMODEL_REQ_PING_DEVICE_STATE_SIG);
                reqEvt->basic.requester = 0;
                reqEvt->deviceId = me->pingSingleDeviceId;
                reqEvt->oneTimePingInitialDeviceId = 0;
                reqEvt->active = TRUE;
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
            }
        }
        return Q_TRAN(me->returnState);
    }
    }
    return Q_SUPER(me->superState);
}

/*..........................................................................*/
static QState SelectB_stopped(SelectB *me, QEvent const *e  __attribute__((unused)))
{
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static void SelectB_SendDsmDoneEvent(DeviceParameterEvt const *reqEvt, uint8_t error)
{
    if (reqEvt->sendDoneModifier == DSM_DONE_NO_EVENT) {
        /* nothing to be done */
        return;
    } else if (reqEvt->sendDoneModifier == EVENT_DSM_DONE_SET_DEVICE_CONFIG) {
        /* send set device config done event */
        DsmApi_SendDsmDoneSetDeviceConfigEvent(error, reqEvt->deviceId, reqEvt->paramBank, reqEvt->paramOffset);
        return;
    }
    /* send set device done event for specified modifier */
    DsmApi_SendDsmDoneSetDeviceEvent(reqEvt->sendDoneModifier, error, reqEvt->deviceId);
}

/*..........................................................................*/
uint32_t SelectB_GetTimeSeconds()
{
    SelectB *me = &l_selectB;
    uint32_t time = 60 * me->minuteCount;
    time += (SELECT_B_ONE_MINUTE_TICKS - me->minuteTimer.ctr) / SELECT_B_ONE_SECOND_TICKS;
    return time;
}

/*..........................................................................*/
void SelectB_sendRestartRegistrationSignal()
{
    static const QEvent registrationEvt = { SELECT_B_RESTART_REGISTRATION, 0 };
    QActive_postFIFO(AO_SelectB, &registrationEvt);
}

/*..........................................................................*/
void SelectB_AnnounceDevice(const char* dsuid, const char* vdc_dsuid, uint32_t message_id)
{
    SelectBAnnounceDeviceEvt* evt = Q_NEW(SelectBAnnounceDeviceEvt, SELECT_B_ANNOUNCE_DEVICE_SIG);
    strncpy(evt->dsuid, dsuid, sizeof(evt->dsuid));
    strncpy(evt->vdcdsuid, vdc_dsuid, sizeof(evt->vdcdsuid));
    evt->message_id = message_id;
    QActive_postFIFO(AO_SelectB, (QEvent *)evt);
}

/*..........................................................................*/
void SelectB_UserProtobufMessageResponse(uint16_t size, const uint8_t* response)
{
    VdcUserProtobufMessageResponseEvt *reqEvt =
        Q_NEW(VdcUserProtobufMessageResponseEvt, VDC_API_USER_PROTOBUF_MESSAGE_RESPONSE_SIG);
    uint16_t pbMessageSize = min(size, sizeof(reqEvt->response));
    reqEvt->size = pbMessageSize;
    memcpy(reqEvt->response, response, pbMessageSize);
    QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
}

